# -*- coding: utf-8 -
#
# This file is part of couchdbkit released under the MIT license.
# See the NOTICE for more information.

version_info = (0, 6, 5)
__version__ =  ".".join(map(str, version_info))
